#!MUNGERPATH

; Copyright (c) 2004, 2005, 2006, James Bailie <jimmy@mammothcheese.ca>.
; All rights reserved.
; 
; Redistribution and use in source form, with or without
; modification, are permitted provided that the following conditions are met:
; 
;     * Redistributions of source code must retain the above copyright
; notice, this list of conditions and the following disclaimer.
;     * The name of James Bailie may not be used to endorse or promote 
; products derived from this software without specific prior written permission.
; 
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS"
; AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
; ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
; LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
; SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
; INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
; CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
; ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
; POSSIBILITY OF SUCH DAMAGE.

; WEB APPLICATION LOGIN HANDLER
; -----------------------------

; This script contains the core functions needed to build a simple web
; application which requires logins for access.  If invoked without CGI
; parameters and without a valid session-id cookie, it presents a login
; form to the user.  After a successful login, it generates a new
; session-id and redirects the client to pull down the script without CGI
; parameters, but with the session-id cookie.  The causes the script to
; invoke the request handler for the main page, which must be associated
; with the string "default" in the requests table.

; Forcing the client to pull down the default page prevents a malicious
; person (a coworker, let's say) from using the browser's back button to
; return to the page after the login page, and then use the refresh button
; to cause the browser to resubmit the login form, and thus gain access to
; the account of a user who has logged-off and then left his computer
; unattended.  The redirect request will replace the form submission in the
; browser's history.  If the malicious person attempts the back-and-reload
; trick on a logged-out account, either the now-invalid previous session-id
; cookie will be presented to this script, which it will reject, or if an
; hour or more has elapsed since the login, the cookie will have expired in
; the browser and not be submitted in the request.  Either way, once
; logged-out, the back-and-reload trick will fail.

; This can easily be circumvented with today's browsers which remember form
; data, and automatically fill-out login forms.  If users insist on having
; their browsers remember their logins in insecure environments, that is
; their problem.  This behavior also ensures that when a user refreshes the
; page after the login page he does not unwittingly login again and restart
; the login timer.

; Successive invocations of the script after login must have the CGI
; "request" parameter set by the caller to indicate what function to call,
; after cookie validation.  The value of the "request" parameter is used
; as a key to lookup the request's handler closure in the requests table.
; Each request handler closure must accept one argument, which will be the
; user name of the login, which it may use however it wishes, or ignore.
; Additionally, a request handler must be defined and bound to the string
; "default" in the requests table.

; The session-id is stored in a sqlite database on the server, and so
; Apache may be stopped and started without affecting the state of active
; logins (They may expire in the meantime, however).  Session-ids expire
; after one hour, when the user must login again to continue accessing the
; application.  The cookie generated at login is also set to expire after
; one hour.

; An example default request handler is shown at the end of the script.
; Once configured for your system, as described below, this script can be
; run as a minimal demonstration program.  When you understand how it
; works, you may alter the value of the 'application variable to cause it
; to load in custom request handlers to override the default one and
; implement your own application.

; If a user logs in while another session-id for that user is still active,
; the new session-id for that user will replace the old one in the
; database.  Any attempt to invoke the script with the old session-id
; cookie will cause the login form to be sent back to the client.
; Therefore only one client may be logged in under any particular user
; name, which is to say each simultaneous user of this script must have a
; unique login.

; To employ this script you need to write your own handlers for the various
; requests your program will recognize.  Your program may employ CGI
; parameters however it likes, but the "request" parameter is reserved for
; use by the login code.  It is the dispatch parameter, and determines
; which function will be invoked to respond to the client.  The decoded
; parameters are available in the "data" table, initialized by the
; "get_data" function of cgi.munger, which is invoked in the
; "service_request" function of this script.

; You should, at a minimum, read the instructions at the top of the
; cgi.munger module (found in (libdir)), and use the functions it provides
; perform encoding-to and decoding-from x-www-form-urlencoding.  The module
; also provides functions to access and create cookies, and send the
; HTTP/1.1 and XHTML headers to the client.

; INSTALLATION
; -----------------------

; The munger interpreter this script runs on must be setuid root to verify
; logins against the user database.  Create a copy of the munger executable
; named suidmunger.  Set the owner to root, and the group to www.  Set the
; permissions to 4050, to turn on the set-user-id bit and allow execution
; only by members of group www.  Verify there are no real users in group www
; in /etc/group, and in the user database itself, or they will be able to
; run suidmunger, completely compromising security.  In addition, if you
; have users who have their own HTML directories inside their home
; directories, Apache should not be configured to run CGI programs from
; those directories, unless you use suexec to run those programs with user
; and group settings other than www.  If you do not do this, your users'
; CGI programs may run suidmunger, completely compromising security.

; This script itself should be owned by www, be in group www, and have
; permissions 0500, to allow only apache to run it.  The eight variables
; below need to be altered to reflect your system.

; ----------------------------------------------------------------------

; Prefix for fatal error messages.  Change it to what you will.

(setq prefix "fatal error: ")

; Hostname of the (possibly virtual) host this script is accessed by.
; This is only used once after login, when we redirect the client
; to pull down the main page.  After that, all URIs this script uses
; are relative URIs.

(setq host "localhost")

; Directory containing this script, either as defined by a ScriptAlias
; directive in httpd.conf, or an actual absolute path.  This value MUST
; BEGIN AND END WITH A SLASH.  If the directory is not defined by a
; ScriptAlias directive, and is not the DocumentRoot directory, it must
; have the ExecCGI option set for it in its Directory directive in
; httpd.conf.

(setq cgi "/cgi-bin/")

; Fully-qualified location in which you would like the logins database to
; reside.  All directory components of this value must already exist before
; the script is run.

(setq logins "/usr/local/www/db/logins.db")

; Path, relative to the location of this script, of the stylesheet the
; login form should reference, or the empty string for no stylesheet.  If
; the stylesheet is in the Apache DocumentRoot directory and this script is
; in another directory, fully-qualify the path using the Apache document
; root directory as the root directory of the path.  For example, use
; "/stylesheet.css" and not "/usr/local/www/data/stylesheet.css".  Have a
; look a the function "login_form" below, to see what xhtml ids have been
; assigned to various components of the form, in order to allow styling
; with CSS.

(setq stylesheet "/login.css")

; The fully-qualified path to the functions implementing your web
; application.  Here, fully-qualified means fully-qualified for your
; filesystem.  Code must be included to map request strings to closures in
; the requests table, for each additional request the script will handle
; beyond "logout".  Be sure to insert a mapping from "default" to your
; default request handler to override the example one.

(setq application "")

; Set to the basename of this script.  This variable allows you to rename
; the script without having to edit anything but this definition.  It is
; essential that this variable be set correctly for the script to function.

(setq script "login.cgi")

; Set to 0 or 1 to indicate whether the script is being run by an SSL
; enabled webserver.  This determines whether the script uses the "http" or
; "https" protocol indicator in its redirection URI, and it will cause
; access to this script via an http URI to generate a redirect to the https
; base URI.  Note that if you do not run this script via an SSL enabled
; server, then users' login data and their session-id cookies will be
; visible to anyone capable of sniffing the packets passing through
; intermediate hosts on the network.

(setq ssl 0)

; Nothing below this line needs to be altered.
; ----------------------------------------------------------------------

; Set to the number of minutes left in login for the request currently
; being processed.

(setq time_left 0)

; Makes lisp errors fatal to the interpreter.

(fatal)

(when (cadr (geteuid))
   (die script ": " prefix "must be root to run"))

; Sets real user to root as well.

(setuid "root")

; Used to prevent SQL injection attacks.

(setq qt_rx (regcomp "'"))

; Loads the cgi.munger module from the library directory.  None of the
; files in this directory are installed with write permissions enabled, and
; they all are originally owned by root and in group wheel.  This should
; prevent malicious users from inserting code here to take control of the
; setuid root Munger interpreter.  For production use, one could make a
; copy of this file and store it somewhere only accessible by group
; "www".

(load (join "/" (libdir) "cgi.munger"))

; Opens the logins database.  Creates the table if the database is being
; created.

(let ((result (exists logins)))

   (when (stringp (setq logins_db (sqlite_open logins)))
      (die script ": " prefix logins_db))

   (unless result
      (let ((sql "CREATE TABLE Logins ( Id TEXT PRIMARY KEY UNIQUE, Time INTEGER, User TEXT )"))
         (when (stringp (setq result (sqlite_exec logins_db sql)))
            (die script ": " prefix result))

         (chown "" "wheel" logins)
         (chmod "600" logins))))

; Peforms logins to database.  Creates cookie with session-id for this
; session.  Both cookie and session-id expire in 60 minutes.

(defun login_to_database (user)
   (letn ((sql (concat "SELECT COUNT(*) FROM Logins WHERE User = '"
                       (substitute qt_rx "''" user 0) "'"))
          (result (sqlite_exec logins_db sql))
          (id (concat (stringify user "-" (random (random 10000)) (time) (random 1000000)))))

      (cond ((stringp result)
             (die script ": " prefix "sqlite_exec: " result))

            ; Replacing a current login:

            ((and result (digitize (car (cadr result))))
             (setq sql (stringify "UPDATE Logins SET Id = '" id "', Time = " (time) " WHERE User = '" user "'"))
             (when (setq result (sqlite_exec logins_db sql))
               (die script ": " prefix "sqlite_exec: " result)))

            ; Entering a new login:

            (1 (setq sql (stringify "INSERT INTO Logins VALUES ( '" id "'," (time) ",'" user "')"))
               (when (setq result (sqlite_exec logins_db sql))
                  (die script ": " prefix "sqlite_exec: " result))))

      ; Create cookie with user id.

      (make_cookie (list (list "id" id)) cgi 3600)))

; Calls checkpass to verify users login data.

(defun login ()
   (let ((user (lookup data "user"))
         (pass (lookup data "pass")))

      (if (or (not user) (not pass) (not (checkpass user pass)))
         (login_form)

         (setq time_left 60)
         (redirect_client (login_to_database user)))))

; Redirects the client to pull down the default page after login to clear
; out the form contents from the browser's history.  We set the session-id
; cookie here, too.

(defun redirect_client ((cookie))
   (println "Date: " (date 1 1 ))

   (when cookie
      (println "Set-Cookie: " (car cookie)))

   (println "Location: http" (if ssl "s" "") "://" host cgi script)
   (newline))

; Sends login form to client.

(defun login_form ()
   (send_headers "Login" stylesheet "" "" "")

   (println "<div id=\"container\">")
   (println "<h1>Login</h1>")
   (println "<form action=\"" cgi script "\" method=\"post\">")
   (println "<input type=\"hidden\" name=\"request\" value=\"login\" />")

   (println "<table id=\"login\">")
   (println "<tr><td class=\"left\"><label for=\"user\">User:</label></td>"
      "<td class=\"right\"><input class=\"text\" type=\"text\" name=\"user\" /></td><tr>")
   (println "<tr><td class=\"left\"><label for=\"pass\">Password:</label></td>"
      "<td class=\"right\"><input class=\"text\" type=\"password\" name=\"pass\" /></td></tr>")
   (println "<tr><td class=\"left\"></td>"
      "<td class=\"right\" id=\"submit\"><input type=\"submit\" value=\"Submit\" /></td></tr>")
   (println "</table>")

   (println "</form>")
   (println "</div>")
   (println "</body>")
   (println "</html>"))

; Validates id in cookie against logins db.

(defun validate_cookie ()
   (let ((id (alist_lookup "id" (get_cookie))))

      (when id
         (letn ((sql (stringify "SELECT User, Time FROM Logins WHERE Id = '"
                                (substitute qt_rx "''" (car id) 0) "'"))
                (result (sqlite_exec logins_db sql)))

            (if (stringp result)
               (die script ": " prefix "sqlite_exec: " result)

               (setq time_left -1)
               (when result
                  (setq time_left (- 60 (/ (timediff (time) (cadr (cadr result))) 60))))

               (if (and (> time_left 0) (<= time_left 60))
                  (car (cadr result))
                  0))))))

; Determines which page to send to client.

(let ((user "")
      (request "")
      (func ""))

   (defun service_request ()
      (get_data)
      (setq request (lookup data "request"))

      ; If we are supposed to be running over SSL and we have not been
      ; invoked by an https URI, then we redirect the client to the base
      ; URI via https.

      (cond ((and ssl (not (getenv "HTTPS")))
             (redirect_client))

            ; The clause below is triggered if there is no request
            ; parameter.  If we receive a valid session-id cookie, then
            ; we send the default page, otherwise we send the login
            ; form.

            ((not request)
             (if (setq user (validate_cookie))
               ((lookup requests "default") user)
               (login_form)))

            ; A login request will not be accompanied by a valid
            ; session-id cookie, because it is generated after
            ; successful login, so this is handled as a special case.
            ; This means "login" does not need an entry in the requests
            ; table.

            ((eq request "login") (login))

            ; Otherwise, if we have a valid session-id cookie, and
            ; a valid request parameter, we invoke the corresponding
            ; request handler.  If not, we send the login form.

            (1 (if (and (setq user (validate_cookie))
                        (setq func (lookup requests request)))
                  (func user)
                  (login_form))))))

; Expire logins older than an hour.

(defun expire_logins ()
   (letn ((t (timethen -3600))
          (sql (stringify "DELETE FROM Logins WHERE Time < " t))
          (result (sqlite_exec logins_db sql)))

      (when (stringp result)
         (die script ": " prefix "sqlite_exec: " result))))

; Explicit logut.

(defun logout (user)
   (letn ((sql (stringify "DELETE FROM Logins WHERE User = '" user "'"))
          (result (sqlite_exec logins_db sql)))

      (when (stringp result)
         (die script ": " prefix "sqlite_exec: " result))

      (login_form)))

; Table mapping requests to request handlers.

(setq requests (table))
(hash requests "logout" logout)

; This is an example default function forming a minimal, but complete login-
; controlled application.

(defun default (user)
   (send_headers "Default Request Handler" stylesheet "" "" "")

   (println "<h1>Example Web Application</h1>")
   (println "<p>Login expires in " time_left " minutes.</p>")
   (println "<p><a href=\"" cgi script "?request=default\">Refresh</a></p>")
   (println "<p><a href=\"" cgi script "?request=logout\">Logout</a></p>")
   (println "</body></html>"))

(hash requests "default" default)

; Loads the function definitions to implement the web application.

(when application
   (load application))

; Removes any expired logins from the database before starting to handle
; requests.

(expire_logins)
(service_request)
(quit)
